import sys
import asyncio
import aiohttp
import os
import os.path
import urllib.parse
import time
import base64
import traceback
import base64
import json
import progressbar

from fastdbfs.fileinfo import FileInfo

class Disconnected():
    def __init__(self):
        pass

    def __getattr__(self, method):
        raise Exception("open must be called first!")

    def prompt(self):
        return "*disconnected* "

class DBFS():

    def __init__(self, id, host, cluster_id, token,
                 workers=8, chunk_size=1048576, max_retries=10,
                 error_delay=10, error_delay_increment=10):
        self.id = id
        self.host = host
        self.cluster_id = cluster_id
        self.token = token
        self.path = "/"
        self.workers = workers
        self.chunk_size = chunk_size
        self.max_retries = max_retries
        self.error_delay = error_delay
        self.error_delay_increment = error_delay_increment
        self.wait_until = 0

        self.semaphore = asyncio.Semaphore(workers)

        # print(f"host: {self.host}, cluster: {self.cluster_id}, token: {self.token}")

    def check(self):
        self._assert_dir(".")

    def _resolve(self, path, *more):
        return os.path.normpath(os.path.join(self.path, path, *more))

    def prompt(self):
        if self.id == "DEFAULT":
            return f"{self.path}$ "
        else:
            return f"{self.id}:{self.path}$ "

    def cd(self, path):
        path = self._resolve(path)
        self._assert_dir(path)
        self.path = path

    async def _unpack_response(self, response):
        status = response.status

        if response.headers.get("x-envoy-ratelimited", None) == "true":
            # print(f"Too many requests!")
            raise RateError("Too many requests")

        if "content-type" not in response.headers:
            print(f"Response headers: {response.headers}")

        ct = response.headers['content-type']
        if ct != "application/json":
            raise Exception(f"Unexpected response received from API, Content-type: {ct}, HTTP status: {status}")

        try:
            data = await response.json()
        except OSError as ex:
            # connection errors and alike should be retried
            raise ex
        except Exception as ex:
            raise Exception("Unable to retrieve API response") from ex

        if status == 200:
            return data

        try:
            error_code = data["error_code"]
            message = data["message"]
        except:
            raise Exception("Unexpected response received from API, HTTP status {status}")

        raise APIError(error_code, message)

    async def _async_control(self, cb):
        retries = 0
        while True:
            async with self.semaphore:
                try:
                    delta = self.wait_until - time.time()
                    if delta > 0:
                        # print(f"Delaying for {delta} seconds")
                        await asyncio.sleep(delta)
                    return await cb()
                except RateError as ex:
                    self.wait_until = time.time() + 1
                    # we do not increase the retries count when a
                    # RateError pops up!
                except OSError as ex:
                    if retries > self.max_retries:
                        raise ex
                    # We keep the semaphore while we sleep in order to
                    # reduce the pressure on the remote side.
                    await asyncio.sleep(self.error_delay + self.error_delay_increment * retries)
                    retries += 1

                except:
                    _, ex, trace = sys.exc_info()
                    print("Stack trace:")
                    traceback.print_tb(trace)
                    raise ex


    async def _async_http_get(self, session, end_point, **params):
        async def cb():
            url = urllib.parse.urljoin(self.host, end_point)
            headers = { "Authorization": "Bearer " + self.token }
            #print(f"session: {session}\nurl: {url}\nparams: {params}\nheaders: {headers}")
            async with session.get(url,
                                   params=params,
                                   headers=headers) as response:
                return await self._unpack_response(response)
        return await self._async_control(cb)


    async def _async_http_post(self, session, end_point, **data):
        async def cb():
            url = urllib.parse.urljoin(self.host, end_point)
            headers = { "Authorization": "Bearer " + self.token,
                        "Content-Type": "application/json" }
            load = json.dumps(data)
            async with session.post(url,
                                    headers=headers,
                                    data=load) as response:
                return await self._unpack_response(response)

        return await self._async_control(cb)

    async def _async_call_with_session(self, cb, *params, **kwparams):
        async with aiohttp.ClientSession() as session:
            return await cb(session, *params, **kwparams)

    async def _async_get_status(self, session, path):
        r = await self._async_http_get(session, "api/2.0/dbfs/get-status",
                                       path=self._resolve(path))
        return FileInfo.from_json(r)

    def _run_with_session(self, task, *args, **kwargs):
        task_with_session = self._async_call_with_session(task, *args, **kwargs)
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(task_with_session)

    def _simple_get(self, end_point, **params):
        return self._run_with_session(self._async_http_get, end_point, **params)

    def _simple_post(self, end_point, **params):
        return self._run_with_session(self._async_http_post, end_point, **params)

    def get_status(self, path):
        return self._run_with_session(self._async_get_status, path)

    async def _async_mkdir(self, session, path):
        path = self._resolve(path)
        await self._async_http_post(session, "api/2.0/dbfs/mkdirs", path=path)

    def mkdir(self, path):
        return self._run_with_session(self._async_mkdir, path=path)

    async def _async_list(self, session, path):
        path = self._resolve(path)
        #print(f"> {path}")
        r = await self._async_http_get(session, "api/2.0/dbfs/list", path=path)
        #print(f"listing {path} done")
        return [FileInfo.from_json(e) for e in r.get("files", [])]

    def ls(self, path):
        path = self._resolve(path)
        fi = self.get_status(path)
        if fi.is_dir():
            return self._run_with_session(self._async_list, path=path)
        else:
            return [fi]

    async def _async_rm(self, session, path, recursive=False):
        path = self._resolve(path)
        await self._async_http_post(session, "api/2.0/dbfs/delete",
                                    path=path, recursive=recursive)

    def rm(self, path, recursive=False):
        path = self._resolve(path)
        self._run_with_session(self._async_rm,
                               path=path, recursive=recursive)

        # Have we removed the CWD?
        if self.path != "/":
            if (self.path == path or
                (self.path.startswith(path) and self.path[len(path)] == "/")):
                # Yes, fix it!
                self.path = os.path.split(path)[0]

    def _assert_dir(self, path):
        if not self.get_status(path).is_dir():
            raise Exception("Not a directory")

    async def _async_create(self, session, path, overwrite=False):
        path = self._resolve(path)
        out = await self._async_http_post(session,
                                     "api/2.0/dbfs/create",
                                     path=path, overwrite=overwrite)
        return int(out["handle"])

    async def _async_add_block(self, session, handle, block):
        return await self._async_http_post(session,
                                           "api/2.0/dbfs/add-block",
                                           data=base64.standard_b64encode(block).decode('ascii'),
                                           handle=handle)

    async def _async_close(self, session, handle):
        return await self._async_http_post(session,
                                           "api/2.0/dbfs/close",
                                           handle=handle)

    async def _async_put_from_file(self, session, infile, target, size=None, overwrite=False, update_cb=None):
        target = self._resolve(target)

        handle = await self._async_create(session, target, overwrite)
        try:
            bytes_copied = 0
            while True:
                chunk = infile.read(self.chunk_size)
                chunk_len = len(chunk)
                if chunk_len == 0:
                    break

                await self._async_add_block(session, handle, chunk)

                bytes_copied += chunk_len
                if update_cb:
                    await update_cb(size, bytes_copied)

            await self._async_close(session, handle)

            fi = await self._async_get_status(session, target)
            if fi.size() != bytes_copied:
                raise Exception(f"corrupted copy detected, copied: {bytes_copied}, remote: {fi.size()}, expected: {size}")
            return fi
        except Exception as ex:
            try: await self._async_rm(target)
            except: pass
            raise ex

    def put_from_file(self, infile, target, size=None, overwrite=False):

        with progressbar.DataTransferBar() as bar:
            if size is not None:
                bar.max_value = size

            async def update_cb(size, bytes_read):
                bar.update(bytes_read)

            return self._run_with_session(self._async_put_from_file,
                                          infile, target, size, overwrite,
                                          update_cb)

    async def _async_put(self, session, src, target, **kwargs):
        size = os.stat(src).st_size
        with open(src, "rb") as infile:
            return await self._async_put_from_file(session, infile, target, size, **kwargs)

    def _new_swarm(self, workers, queue_max_size, use_priority_queue=False):

        queue_class = asyncio.PriorityQueue if use_priority_queue else asyncio.Queue
        # print(f"allocating queue of class {queue_class}")
        queue = queue_class(maxsize = queue_max_size)

        async def worker(ix):
            # print(f"worker {ix} started")

            async with aiohttp.ClientSession() as session:
                while True:
                    # print(f"worker {ix}  waiting for tasks")
                    (task_key, cb, response_queue, kwargs) = await queue.get()
                    if cb is None:
                        return
                    try:
                        # print(f"cb kwargs: {kwargs}")
                        value = await cb(session, **kwargs)
                        res = (task_key, value, None)
                    except Exception as ex:
                        #_, ex, trace = sys.exc_info()
                        #print("Stack trace:")
                        #traceback.print_tb(trace)

                        res = (task_key, None, ex)
                    #print(f"Queueing response for {task_key}")
                    if response_queue:
                        await response_queue.put(res)

        async def end():
            for ix in range(workers):
                await self._queue_task(queue, None, task_key=ix)

        swarm = asyncio.gather(*[worker(ix) for ix in range(workers)])
        return (swarm, queue, end)

    async def _queue_task(self, _queue, _task,
                          task_key=None, response_queue=None,
                          **kwargs):
        return await _queue.put((task_key, _task, response_queue, kwargs))

    def _unwrap_response(self, res):
        (_, value, ex) = res
        if ex is not None:
            raise ex
        return value

    async def _async_get_chunk(self, session, path, offset, length, out):
        # print(f"processing chunk at {offset}")
        remaining = length
        while remaining > 0:
            r = await self._async_http_get(session,
                                           "api/2.0/dbfs/read",
                                           path=path, offset=offset, length=length)

            bytes_read = r["bytes_read"]
            if bytes_read == 0 or bytes_read > remaining:
                raise Exception("Invalid data response")

            chunk = base64.standard_b64decode(r["data"])
            if len(chunk) != bytes_read:
                raise Exception("Invalid data response, size is not as promised")

            out.seek(offset)
            out.write(chunk)
            remaining -= bytes_read
            offset += bytes_read
        # print(f"chunk of {length} bytes copied at offset {offset - length}")
        return length

    async def _async_get(self, session, low_queue, src, target,
                         overwrite=False, update_cb=None):
        # FIXME: a race condition exists here!
        if not overwrite and os.path.exists(target):
            raise Exception("File already exists")
        with open(target, "wb") as out:
            await self._async_get_to_file(session, low_queue, src, out, update_cb)

    async def _async_get_to_file(self, session, low_queue, src, out, update_cb=None):
        src = self._resolve(src)
        fi = await self._async_get_status(session, src)
        size = fi.size()

        response_queue = asyncio.Queue()
        active_tasks = 0 # we count the request we have queued
        bytes_copied = 0
        offset = 0

        try:
            while True:
                while True:
                    try:
                        res = response_queue.get_nowait()
                    except asyncio.QueueEmpty:
                        break
                    active_tasks -= 1
                    bytes_copied += self._unwrap_response(res)
                    if update_cb:
                        await update_cb(size, bytes_copied)

                next_offset = min(offset + self.chunk_size, size)
                length = next_offset - offset
                if length <= 0:
                    break

                await self._queue_task(low_queue,
                                       self._async_get_chunk,
                                       response_queue=response_queue,
                                       path=src, offset=offset, length=length,
                                       out=out)
                active_tasks += 1
                offset = next_offset

            while active_tasks > 0:
                res = await response_queue.get()
                active_tasks -= 1
                bytes_copied += self._unwrap_response(res)
                if update_cb:
                    await update_cb(size, bytes_copied)

        finally:
            for _ in range(active_tasks):
                try: await response_queue.get()
                except: pass

        return fi


    def get_to_file(self, src, out):

        (swarm, queue, end) = self._new_swarm(self.workers, self.workers * 2)

        with progressbar.DataTransferBar() as bar:
            async def update_cb(size, read):
                bar.max_value = size
                bar.update(read)


            result = None
            async def control():
                nonlocal result
                try:
                    result = await self._async_call_with_session(self._async_get_to_file,
                                                                 queue, src, out, update_cb)
                finally:
                    await end()

            loop = asyncio.get_event_loop()
            loop.run_until_complete(asyncio.gather(control(), swarm))
            return result

    async def _async_find(self, session, path, update_cb):
        path = self._resolve(path)
        root_fi = await self._async_get_status(session, path)

        if not root_fi.is_dir():
            await update_cb(root_fi, True, None)
            return

        response_queue = asyncio.Queue()
        (swarm, queue, end) = self._new_swarm(workers = self.workers,
                                              queue_max_size = self.workers,
                                              use_priority_queue=True)

        async def control():
            await self._queue_task(queue,
                                   self._async_list,
                                   task_key = path,
                                   response_queue = response_queue,
                                   path = path)

            pending = [[path, root_fi, None, None]] # path, fileinfo, ok, exception

            while pending:
                (key, fis, ex) = await response_queue.get()
                # update the entry
                for e in pending:
                    if e[0] == key:
                        if ex:
                            e[2] = False
                            e[3] = ex
                        else:
                            e[2] = True
                        break
                else:
                    raise Exception("Internal error, key not found in pending")

                if fis:
                    # add the new entries
                    pending += [[fi.abspath(), fi,
                                 None if fi.is_dir() else True,
                                 None]
                                for fi in fis]
                    pending.sort()

                    # request their listings
                    for fi in fis:
                        if fi.is_dir():
                            #print(f"< {fi.abspath()}")
                            await self._queue_task(queue,
                                                   self._async_list,
                                                   task_key = fi.abspath(),
                                                   response_queue = response_queue,
                                                   path = fi.abspath())

                # finally report entries ready from the top
                while pending and pending[0][2] is not None:
                    head = pending.pop(0)
                    await update_cb(head[1], head[2], head[3])
            await end()
        await asyncio.gather(control(), swarm)

    def find(self, path, update_cb):
        path = self._resolve(path)

        async def _async_update_cb(fi, ok, ex):
            update_cb(fi, ok, ex)

        self._run_with_session(self._async_find,
                               path,
                               _async_update_cb)

    async def _async_rget(self, session, src, target, update_cb=None):
        src = self._resolve(src)

        # We use two worker queues here, low is for the low level
        # requests. We pass it to the _self_get method without ever
        # using it directly. high_queue is for parallelizing the
        # _self_get calls.

        (high_swarm, high_queue, high_end) = self._new_swarm(self.workers, self.workers * 2)
        (low_swarm, low_queue, low_end) = self._new_swarm(self.workers, self.workers * 2)
        response_queue = asyncio.Queue()
        active_gets = 0

        async def empty_response_queue():
            nonlocal active_gets
            while True:
                try:
                    (relpath, res, ex) = response_queue.get_nowait()
                except asyncio.QueueEmpty:
                    break
                active_gets -= 1
                ok = ex is None
                await update_cb(relpath, ok, ex)

        async def find_cb(fi, ok, ex):
            nonlocal active_gets

            # We empty the response_queue everytime this function is called.
            await empty_response_queue()

            # print(f"entry found: {fi.abspath()}, ok: {ok}, ex: {ex}")

            remote_path = fi.abspath()
            relpath = fi.relpath(src)
            local_path = os.path.abspath(os.path.join(target, relpath))

            if fi.is_dir():
                try:
                    # print(f"creating local path {local_path}")
                    os.mkdir(local_path)
                except Exception as ex:
                    if not os.path.isdir(local_path):
                        if update_cb:
                            await update_cb(relpath, False, ex)
            else:
                await self._queue_task(high_queue,
                                       self._async_get,
                                       task_key=relpath,
                                       response_queue=response_queue,
                                       low_queue=low_queue, src=remote_path, target=local_path)
                active_gets += 1

        # So, here we go!
        # First we run the find task...
        await self._async_find(session, src, find_cb)

        # Then we finish the high level queue...
        await high_end()

        # Remove any pending notification from response_queue...
        while active_gets > 0:
            (relpath, res, ex) = await response_queue.get()
            active_gets -= 1
            ok = ex is None
            await update_cb(relpath, ok, ex)

        # And finally terminate the low queue...
        await low_end()


    def rget(self, src, target, update_cb=None):
        async def _async_update_cb(*args):
            update_cb(*args)
        self._run_with_session(self._async_rget,
                               src, target,
                               _async_update_cb if update_cb else None)

    async def _async_rput(self, session, src, target, update_cb=None):
        target = self._resolve(target)
        local_root_fi = FileInfo.from_local(src)

        (swarm, queue, end) = self._new_swarm(self.workers, self.workers * 2)
        response_queue = asyncio.Queue()

        async def control():
            active_puts = 0
            for (dir, _, filenames) in os.walk(src):
                local_fi = FileInfo.from_local(dir)
                relpath = local_fi.relpath(local_root_fi.abspath())
                remote_path = self._resolve(target, relpath)
                local_path = os.path.join(src, relpath)

                status = None
                try:
                    await self._async_mkdir(session, path=remote_path)
                    if update_cb:
                        await update_cb(local_path, True, None)

                except Exception as ex:
                    if update_cb:
                        await update_cb(local_path, False, ex)
                        for fn in filenames:
                            await update_cb(os.path.join(local_path, fn), False, None)

                else:
                    for fn in filenames:
                        local_child_path = os.path.join(local_path, fn)
                        remote_child_path = self._resolve(remote_path, fn)
                        #print(f"remote_child_path: {remote_child_path}")
                        await self._queue_task(queue,
                                               self._async_put,
                                               task_key=local_child_path,
                                               src=local_child_path,
                                               target=remote_child_path,
                                               response_queue=response_queue)
                        active_puts += 1


                while True:
                    try:
                        (local_child_path, res, ex) = response_queue.get_nowait()
                    except asyncio.QueueEmpty:
                        break
                    active_puts -= 1
                    if update_cb:
                        ok = ex is None
                        await update_cb(local_child_path, ok, ex)

            while active_puts > 0:
                (local_child_path, res, ex) = await response_queue.get()
                active_puts -= 1
                if update_cb:
                    ok = ex is None
                    await update_cb(local_child_path, ok, ex)

            await end()
        await asyncio.gather(control(), swarm)

    def rput(self, src, target, update_cb=None):
        async def _async_update_cb(fi, ok, ex):
            update_cb(fi, ok, ex)
        self._run_with_session(self._async_rput, src, target,
                               _async_update_cb if update_cb else None)




import stat

import os
import os.path

class FileInfo():

    @staticmethod
    def from_json(json):
        return FileInfo(json["is_dir"],
                        json["file_size"],
                        json["modification_time"],
                        json["path"])

    @staticmethod
    def from_local(filename, *more):
        path = os.path.abspath(os.path.join(filename, *more))
        st = os.stat(path)
        return FileInfo(stat.S_ISDIR(st.st_mode),
                        st.st_size,
                        st.st_mtime,
                        path)

    def __init__(self, is_dir, size, mtime, abspath):
        self._is_dir = is_dir
        self._size = size
        self._mtime = mtime
        self._abspath = abspath

    def basename(self):
        return os.path.basename(self._abspath)

    def abspath(self):
        return self._abspath

    def relpath(self, base):
        base = os.path.normpath(base)
        path = self.abspath()
        if path == base:
            return "."
        if path.startswith(base):
            if base == "/":
                return path[1:]
            elif path[len(base)] == "/":
                return path[(len(base)+1):]
        return path

    def is_dir(self):
        return self._is_dir

    def mtime(self):
        return self._mtime

    def size(self):
        return self._size

    def type(self):
        return "dir" if self._is_dir else "file"




import sys
import cmd
import configparser
import os
import os.path
import traceback
import time
import shlex
import pathlib
import subprocess
import tempfile

from fastdbfs.dbfs import DBFS, Disconnected

class CLI(cmd.Cmd):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._init_cfg()
        self._dbfs = Disconnected()
        self._set_prompt()
        self._debug = True

    def cfg(self, section, key, default=None):
        try:
            return self._cfg[section][key]
        except:
            if default is not None:
                return default
            raise Exception(f"Configuration entry {key} missing in section {section}")

    def cfg_int(self, section, key):
        v = self.cfg(section, key)
        try:
            return int(v)
        except:
            raise Exception(f"Configuration entry {key} is not an integer")

    def _init_cfg(self):
        self._cfg = configparser.ConfigParser()
        self._cfg["fastdbfs"] = {
            "workers": 8,
            "chunk_size": 1048576,
            "max_retries": 10,
            "pager": "less",
            "error_delay": 10,
            "error_delay_increment": 10
        }

        home = os.path.expanduser("~")
        fns = [os.path.join(home, fn) for fn in (".databrickscfg", ".fastdbfs", ".config/fastdbfs")]
        self._cfg.read(fns)

    def _tell_error(self, msg):
        _, ex, trace = sys.exc_info()
        print(f"{msg}: {ex}")
        if self._debug:
            print("Stack trace:")
            traceback.print_tb(trace)

    def do_open(self, arg):
        try:
            id, = self._parse(arg, "DEFAULT")
            dbfs = DBFS(id,
                        host = self.cfg(id, "host"),
                        cluster_id = self.cfg(id, "cluster_id"),
                        token = self.cfg(id, "token"),
                        chunk_size = self.cfg_int("fastdbfs", "chunk_size"),
                        workers = self.cfg_int("fastdbfs", "workers"),
                        max_retries = self.cfg_int("fastdbfs", "max_retries"))

            dbfs.check()
            self._dbfs = dbfs
        except:
            self._tell_error(f"Unable to open {arg}")

    def do_cd(self, arg):
        try:
            path, = self._parse(arg, "/")
            self._dbfs.cd(path)
        except:
            self._tell_error(f"{path}: unable to change dir")

    def do_mkcd(self, arg):
        self.do_mkdir(arg)
        self.do_cd(arg)

    def _format_size(self, size):
        if (size >= 1073741824):
            return "%.1fG" % (size / 1073741824)
        if (size > 1024 * 1024):
            return "%.1fM" % (size / 1048576)
        if (size > 1024):
            return "%.1fK" % (size / 1024)
        return str(int(size))

    def _format_time(self, mtime):
        return time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(mtime/1000))

    def do_mkdir(self, arg):
        try:
            path, = self._parse(arg, min=1)
            self._dbfs.mkdir(path)
        except:
            self._tell_error(f"{path}: mkdir failed")

    def do_ls(self, arg):
        try:
            path, = self._parse(arg, ".")
            # cols are type, size, mtime and path
            type_len = 3
            size_len = 1
            mtime_len = 1
            table = []
            for e in self._dbfs.ls(path):
                row = (e.type(),
                       self._format_size(e.size()),
                       self._format_time(e.mtime()),
                       e.basename())

                # print(f"row: {row}")
                type_len = max(type_len, len(row[0]))
                size_len = max(size_len, len(row[1]))
                mtime_len = max(mtime_len, len(row[2]))
                table.append(row)

            fmt = "{:>"+str(type_len)+"} {:>"+str(size_len)+"} {:>"+str(mtime_len)+"} {}"
            for e in table:
                print(fmt.format(*e))

        except:
            self._tell_error(f"{arg}: unable to list directory")

    def do_lcd(self, arg):
        try:
            path, = self._parse(arg, os.path.expanduser("~"))
            os.chdir(path)
        except:
            self._tell_error(f"{arg}: unable to change dir")

    def do_lpwd(self, arg):
        self._parse(arg)
        try:
            print(os.getcwd())
        except:
            self._tell_error("getcwd failed")

    def _local_mkdir(self, path):
        pathlib.Path(path).mkdir(parents=True, exist_ok=True)

    def do_put(self, arg):
        try:
            src, target = self._parse(arg, ".", min=1)
            try:
                # We check whether the target exists and if it is a
                # directory.  If it is a directory we compose the name
                # using the src basename and check the target again.
                for first in (True, False):
                    fi = self._dbfs.get_status(target)
                    if first and fi.is_dir():
                        target = os.path.join(target, os.path.basename(src))
                    else:
                        break
            except Exception as ex:
                print(f"file not found at {target}: {ex}")
                # No file there, ok!
                pass
            else:
                raise Exception("File already exists")

            print(f"copying to {target}")

            size = os.stat(src).st_size

            with open(src, "rb") as infile:
                start = time.time()
                self._dbfs.put_from_file(infile, target, size=size)
                delta = max(1, time.time() - start)

        except:
            self._tell_error(f"{arg}: put failed")

    def _get_to_temp(self, src, prefix=".tmp-", suffix=None, **kwargs):
        try:
            if suffix is None:
                bn = os.path.basename(src)
                try: suffix = bn[bn.rindex("."):]
                except: suffix = ""

            (f, target) = tempfile.mkstemp(prefix=prefix, suffix=suffix, **kwargs)
            out = os.fdopen(f, "wb")
            self._dbfs.get_to_file(src, out)
            out.close()
            return target

        except Exception as ex:
            try: out.close()
            except: pass
            try: os.remove(tmp_fn)
            except: pass
            raise ex

    def do_get(self, arg):
        try:
            src, target = self._parse(arg, ".", min=1)

            if os.path.isdir(target):
                target = os.path.join(target, os.path.basename(src))
            if os.path.exists(target):
                raise Exception("file already exists")

            parent_dir, _ = os.path.split(target)
            self._local_mkdir(parent_dir)

            tmp_target = self._get_to_temp(src, prefix=".transferring-", suffix="", dir=parent_dir)
            os.rename(tmp_target, target)

        except:
            self._tell_error(f"{arg}: unable to retrieve remote file")

    def _get_and_call(self, src, cb):
        target = self._get_to_temp(src)
        try:
            cb(target)
        finally:
            try: os.remove(target)
            except: pass

    def do_rm(self, arg):
        try:
            path, = self._parse(arg, min=1)
            self._dbfs.rm(path)
            print("File removed");
        except:
            self._tell_error(f"{arg}: unable to remove remote file")

    def do_rmdir(self, arg):
        try:
            path, = self._parse(arg, min=1)
            self._dbfs.rm(path, recursive=True)
            print("Dir removed");
        except:
            self._tell_error(f"{arg}: unable to remove remote file")

    def do_find(self, arg):
        try:
            path, = self._parse(arg, ".")
            path = self._dbfs._resolve(path)

            def update_cb(fi, ok, ex):
                print("{:>4} {:>7} {:>19} {}".format(fi.type(),
                                                     self._format_size(fi.size()),
                                                     self._format_time(fi.mtime()),
                                                     fi.relpath(path)))
                if ex:
                    print("# Unable to recurse into {fi.relpath(path)}, {ex}")
                    raise ex

            self._dbfs.find(path, update_cb)

        except:
            self._tell_error(f"{arg}: unable to recursively list remote dir");

    def _rgetput_update_cb(self, path, ok, ex):
        if ok:
            print(f"{path} ok!")
        elif ex is None:
            print(f"{path} FAILED!")
        else:
            print(f"{path} FAILED {ex}!")

    def do_rput(self, arg):
        try:
            src, target = self._parse(arg, None, min=1)
            normalized_src = os.path.normpath(src)
            if target is None:
                if normalized_src == "." or normalized_src == "/":
                    target = "."
                else:
                    target = os.path.basename(normalized_src)

            self._dbfs.rput(src, target, self._rgetput_update_cb)
        except:
            self._tell_error(f"{arg}: rput failed")

    def do_rget(self, arg):
        try:
            src, target = self._parse(arg, None, min=1)
            normalized_src = os.path.normpath(src)
            if target is None:
                if normalized_src == "." or normalized_src == "/":
                    target = "."
                else:
                    target = os.path.basename(normalized_src)

            self._dbfs.rget(src, target, self._rgetput_update_cb)
        except:
            self._tell_error(f"{arg}: rget failed")

    def do_cat(self, arg):
        try:
            src, = self._parse(arg, min=1)
            def cb(fn):
                subprocess.run(["cat", "--", fn])
            self._get_and_call(arg, cb)
        except:
            self._tell_error(f"{arg}: unable to show file")

    def do_more(self, arg, pager=None):
        try:
            src, = self._parse(arg, min=1)
            if pager is None:
                pager = self.cfg("fastdbfs", "pager")
            def cb(fn):
                subprocess.run([pager, "--", fn])
            self._get_and_call(arg, cb)
        except:
            self._tell_error(f"{arg}: unable to show file")

    def do_edit(self, arg, editor=None):
        try:
            src, = self._parse(arg, min=1)
            if editor is None:
                editor = self.cfg("fastdbfs", "editor", default=os.environ.get("EDITOR", "vi"))

            def cb(tmp_fn):
                # in order to avoid race conditions we force the mtime
                # of the temporal file into the past
                the_past = int(time.time() - 2)
                os.utime(tmp_fn, times=(the_past, the_past))
                subprocess.run([editor, "--", tmp_fn])
                stat_after = os.stat(tmp_fn)
                if (stat_after.st_mtime > the_past):
                    with open(tmp_fn, "rb") as infile:
                        self._dbfs.put_from_file(infile, src, size=stat_after.st_size, overwrite=True)
                else:
                    raise Exception(f"File was not modified!")

            self._get_and_call(arg, cb)
        except:
            self._tell_error(f"{arg}: edit failed")

    def do_mg(self, arg):
        self.do_edit(arg, editor="mg")

    def do_vi(self, arg):
        self.do_edit(arg, editor="vi")

    def do_less(self, arg):
        self.do_more(arg, pager="less")

    def do_batcat(self, arg):
        self.do_more(arg, pager="batcat")

    def do_shell(self, arg):
        os.system(arg)

    def do_EOF(self, arg):
        print("\nBye!")
        return True

    def do_exit(self, arg):
        print("Bye!")
        return True

    def _set_prompt(self):
        if self._dbfs:
            self.prompt = self._dbfs.prompt()
        else:
            self.prompt = "*disconnected* "

    def postcmd(self, stop, line):
        self._set_prompt()
        return stop

    def preloop(self):
        if self.cfg("DEFAULT", "token", None) is not None:
            self.do_open("")
        self._set_prompt()

    def _parse(self, arg, *defaults, min=0, max=None):
        args = shlex.split(arg)
        max1 = max if max is not None else min + len(defaults)
        if ((len(args) < min) or
            (max1 >= 0 and len(args) > max1)):
            raise Exception("wrong number of arguments")

        args += defaults[len(args)-min:]

        return args



#!/usr/bin/env python3

from setuptools import setup

import os
pkg_dir = os.path.abspath(os.path.dirname(__file__))
readme_fn = os.path.join(pkg_dir, 'README.md')
with open(readme_fn, encoding='utf-8') as file:
    long_description = file.read()

setup(
    name='fastdbfs',
    version='0.2',
    description="Interactive command line client for Databricks DBFS",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="http://github.com/salva/fastdbfs",
    author="Salvador Fandiño García",
    author_email="sfandino@yahoo.com",
    license='GPLv3+',
    license_files = ('LICENSE.txt',),
    packages=['fastdbfs'],
    install_requires=[ "progressbar2", "aiohttp" ],
    entry_points={'console_scripts': ['fastdbfs=fastdbfs.runner:run']}
);
